import datetime
from contextlib import contextmanager

from .constants import ObservationZoneType


class Writer:
    """
    A writer class for the `XCSoar <http://www.xcsoar.org>`_ task file format::

        with open('task.tsk', 'w') as fp:
            writer = Writer(fp)

    This task file format contains one task per file. Writing more than one
    task will cause an exception. A task can be written by using the
    :meth:`~aerofiles.xcsoar.Writer.write_task` method.
    """

    def __init__(self, fp, encoding='utf-8'):
        self.fp = fp
        self.encoding = encoding
        self.indent_level = 0

    def write_line(self, line):
        indent = '\t' * self.indent_level
        self.fp.write((indent + line + '\n').encode(self.encoding))

    def format_tag_content(self, _name, **kw):
        params = list(map(lambda item: '%s="%s"' % item, kw.items()))
        params.sort()
        return _name + ' ' + ' '.join(params)

    def convert_bool(self, kw, key):
        if key in kw:
            value = kw[key]
            if isinstance(value, bool):
                kw[key] = (1 if value else 0)

    def convert_time(self, kw, key):
        if key in kw:
            value = kw[key]
            if isinstance(value, datetime.time):
                kw[key] = value.strftime('%H:%M')

    def convert_timedelta(self, kw, key):
        if key in kw:
            value = kw[key]
            if isinstance(value, datetime.timedelta):
                kw[key] = value.seconds

    @contextmanager
    def write_tag_with_content(self, _name, **kw):
        self.write_line('<%s>' % self.format_tag_content(_name, **kw))
        self.indent_level += 1
        yield
        self.indent_level -= 1
        self.write_line('</%s>' % _name)

    def write_tag(self, _name, **kw):
        self.write_line('<%s/>' % self.format_tag_content(_name, **kw))

    def write_task(self, **kw):
        """
        Write the main task to the file::

            with writer.write_task(type=TaskType.RACING):
                ...

            # <Task type="RT"> ... </Task>

        Inside the with clause the :meth:`~aerofiles.xcsoar.Writer.write_point`
        method should be used to write the individual task points. All
        parameters are optional.

        :param type: type of the task (one of the constants in
            :class:`~aerofiles.xcsoar.constants.TaskType`)
        :param start_requires_arm: ``True``: start has to be armed manually,
            ``False``: task will be started automatically
        :param start_max_height: maximum altitude when the task is started
            (in m)
        :param start_max_height_ref: altitude reference of
            ``start_max_height`` (one of the constants in
            :class:`~aerofiles.xcsoar.constants.AltitudeReference`)
        :param start_max_speed: maximum speed when the task is started (in m/s)
        :param start_open_time: time that the start line opens as
            :class:`datetime.time`
        :param start_close_time: time that the start line is closing as
            :class:`datetime.time`
        :param aat_min_time: AAT time as :class:`datetime.timedelta`
        :param finish_min_height: minimum altitude when the task is finished
            (in m)
        :param finish_min_height_ref: altitude reference of
            ``finish_min_height`` (one of the constants in
            :class:`~aerofiles.xcsoar.constants.AltitudeReference`)
        :param fai_finish: ``True``: FAI finish rules apply
        """

        self.convert_timedelta(kw, 'aat_min_time')
        self.convert_time(kw, 'start_open_time')
        self.convert_time(kw, 'start_close_time')
        self.convert_bool(kw, 'fai_finish')
        self.convert_bool(kw, 'start_requires_arm')

        return self.write_tag_with_content('Task', **kw)

    def write_point(self, **kw):
        """
        Write a task point to the file::

            with writer.write_point(type=PointType.TURN):
                writer.write_waypoint(...)
                writer.write_observation_zone(...)

            # <Point type="Turn"> ... </Point>

        Inside the with clause the
        :meth:`~aerofiles.xcsoar.Writer.write_waypoint` and
        :meth:`~aerofiles.xcsoar.Writer.write_observation_zone` methods must be
        used to write the details of the task point.

        :param type: type of the task point (one of the constants in
            :class:`~aerofiles.xcsoar.constants.PointType`)
        """

        assert 'type' in kw

        self.convert_bool(kw, 'score_exit')

        return self.write_tag_with_content('Point', **kw)

    def write_waypoint(self, **kw):
        """
        Write a waypoint to the file::

            writer.write_waypoint(
                name='Meiersberg',
                latitude=51.4,
                longitude=7.1
            )

            # <Waypoint name="Meiersberg">
            #     <Location latitude="51.4" longitude="7.1"/>
            # </Waypoint>

        :param name: name of the waypoint
        :param latitude: latitude of the waypoint (in WGS84)
        :param longitude: longitude of the waypoint (in WGS84)
        :param altitude: altitude of the waypoint (in m, optional)
        :param id: internal id of the waypoint (optional)
        :param comment: extended description of the waypoint (optional)
        """

        assert 'name' in kw
        assert 'latitude' in kw
        assert 'longitude' in kw

        location_kw = {
            'latitude': kw['latitude'],
            'longitude': kw['longitude'],
        }

        del kw['latitude']
        del kw['longitude']

        with self.write_tag_with_content('Waypoint', **kw):
            self.write_tag('Location', **location_kw)

    def write_observation_zone(self, **kw):
        """
        Write an observation zone declaration to the file::

            writer.write_observation_zone(
                type=ObservationZoneType.CYLINDER,
                radius=30000,
            )

            # <ObservationZone type="Cylinder" radius="30000"/>

        The required parameters depend on the type parameter. Different
        observation zone types require different parameters.

        :param type: observation zone type (one of the constants in
            :class:`~aerofiles.xcsoar.constants.ObservationZoneType`)

        :param length: length of the line
            (only used with type
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.LINE`)
        :param radius: (outer) radius of the observation zone
            (used with types
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.CYLINDER`,
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.SECTOR`,
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.SYMMETRIC_QUADRANT` and
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.CUSTOM_KEYHOLE`)
        :param inner_radius: inner radius of the observation zone
            (only used with type
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.CUSTOM_KEYHOLE`)
        :param angle: angle of the observation zone
            (only used with type
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.CUSTOM_KEYHOLE`)
        :param start_radial: start radial of the observation zone
            (only used with type
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.SECTOR`)
        :param end_radial: end radial of the observation zone
            (only used with type
            :const:`~aerofiles.xcsoar.constants.ObservationZoneType.SECTOR`)
        """

        assert 'type' in kw

        if kw['type'] == ObservationZoneType.LINE:
            assert 'length' in kw

        elif kw['type'] == ObservationZoneType.CYLINDER:
            assert 'radius' in kw

        elif kw['type'] == ObservationZoneType.SECTOR:
            assert 'radius' in kw
            assert 'start_radial' in kw
            assert 'end_radial' in kw

        elif kw['type'] == ObservationZoneType.SYMMETRIC_QUADRANT:
            assert 'radius' in kw

        elif kw['type'] == ObservationZoneType.CUSTOM_KEYHOLE:
            assert 'radius' in kw
            assert 'inner_radius' in kw
            assert 'angle' in kw

        self.write_tag('ObservationZone', **kw)



import math


class Writer:

    """
    A higher-level writer for the OpenAir airspace file format::

        with open('airspace.txt', 'wb') as fp:
            writer = Writer(fp)

    :param fp: file pointer to write to
    :param encoding: the encoding used for the output

    see `OpenAir file format specification
    <http://www.winpilot.com/UsersGuide/UserAirspace.asp>`_

    This class should be used to write records as described under Reader
    into a file.

        writer.write_record(record)

    Currently only airspace records are implemented. Terrain is missing.

    """

    def reset_V(self):
        self.V_X = None
        self.V_D = 1

    def __init__(self, fp=None, encoding='utf-8'):
        self.fp = fp
        self.encoding = encoding
        self.reset_V()

    def format_dms(self, decimal_degrees):
        decimals, number = math.modf(decimal_degrees)
        deg = int(number)
        mnt = int(decimals * 60)
        sec = round((decimal_degrees - deg - mnt / 60) * 3600.00)
        if sec == 60:
            sec = 0
            mnt += 1
            if mnt == 60:
                mnt = 0
                deg += 1
        return deg, mnt, sec

    def format_degree(self, v, width):
        (degrees, minutes, seconds) = self.format_dms(abs(v))
        result = "%0*d:%02d:%02d" % (width, degrees, minutes, seconds)
        return result

    def format_coord(self, P):
        result = self.format_degree(abs(P[0]), 2) + " "
        if P[0] >= 0:
            result = result + "N "
        else:
            result = result + "S "

        result = result + self.format_degree(abs(P[1]), 3) + " "
        if P[1] >= 0:
            result = result + "E"
        else:
            result = result + "W"
        return result

    def write_line(self, line):
        self.fp.write((line + u'\r\n').encode(self.encoding))

    def write_V_X(self, center):
        if center != self.V_X:
            self.V_X = center
            center = self.format_coord(center)
            self.write_line('V X=' + center)

    def write_V_D(self, clockwise):
        if clockwise != self.V_D:
            self.V_D = clockwise
            if clockwise:
                D = "+"
            else:
                D = "-"
            self.write_line('V D=' + D)

    def write_DC(self, element):
        self.write_V_X(element["center"])
        self.write_line('DC %s' % (str(element["radius"])))

    def write_DA(self, element):
        self.write_V_X(element["center"])
        self.write_V_D(element["clockwise"])
        self.write_line('DA %s,%s,%s' % (str(element["radius"]), str(element["start"]), str(element["end"])))

    def write_DB(self, element):
        self.write_V_X(element["center"])
        self.write_V_D(element["clockwise"])
        start = self.format_coord(element["start"])
        end = self.format_coord(element["end"])
        self.write_line('DB %s, %s' % (start, end))

    def write_DP(self, element):
        location = self.format_coord(element["location"])
        self.write_line('DP %s' % (location))

    def write_airspace_element(self, element):
        if element["type"] == "point":
            self.write_DP(element)
        elif element["type"] == "circle":
            self.write_DC(element)
        elif element["type"] == "arc":
            if "radius" in element:
                self.write_DA(element)
            else:
                self.write_DB(element)

    def write_airspace(self, record):
        self.reset_V()
        self.write_line('AC ' + record["class"])
        if "ident" in record and record["ident"]:
            self.write_line('AI ' + record["ident"])
        if "airspace_type" in record and record["airspace_type"]:
            self.write_line('AY ' + record["airspace_type"])
        self.write_line('AN ' + record["name"])
        self.write_line('AH ' + record["ceiling"])
        self.write_line('AL ' + record["floor"])
        if "ground_name" in record and record["ground_name"]:
            self.write_line('AG ' + record["ground_name"])
        if "freq" in record and record["freq"]:
            self.write_line('AF ' + record["freq"])
        element_prev = None
        for element in record["elements"]:
            if element_prev != element:
                element_prev = element
                self.write_airspace_element(element)

    def write_record(self, record):
        if record["type"] == "airspace":
            self.write_airspace(record)
        else:
            raise ValueError('unknown record type: ' + record["type"])



class TaskType:
    AAT = 'AAT'
    FAI_GENERAL = 'FAIGeneral'
    FAI_GOAL = 'FAIGoal'
    FAI_OUT_AND_RETURN = 'FAIOR'
    FAI_TRIANGLE = 'FAITriangle'
    MAT = 'MAT'
    MIXED = 'Mixed'
    RACING = 'RT'
    TOURING = 'Touring'


class PointType:
    START = 'Start'
    OPTIONAL_START = 'OptionalStart'
    AREA = 'Area'
    TURN = 'Turn'
    FINISH = 'Finish'


class ObservationZoneType:
    BGA_START = 'BGAStartSector'
    BGA_FIXED = 'BGAFixedCourse'
    BGA_ENHANCED = 'BGAEnhancedOption'
    CUSTOM_KEYHOLE = 'CustomKeyhole'
    CYLINDER = 'Cylinder'
    FAI_SECTOR = 'FAISector'
    KEYHOLE = 'Keyhole'
    LINE = 'Line'
    MAT_CYLINDER = 'MatCylinder'
    SECTOR = 'Sector'
    SYMMETRIC_QUADRANT = 'SymmetricQuadrant'


class AltitudeReference:
    AGL = 'AGL'
    MSL = 'MSL'



import datetime


class TimeZoneFix(datetime.tzinfo):
    """
    A simple implementation of tzinfo to store timezone of IGC fixes.

    An alternative would be, to use "timezone" which is part of
    Python since v3.2. However, because we try to be python 2.6+
    compatible, we use this implementation. It can be replaced,
    if we set Python 3.x as a minimum requirement for aerofiles.
    """

    def __init__(self, fix=0):
        self.fix = fix

    def __eq__(self, other):
        if isinstance(other, TimeZoneFix):
            return self.fix == other.fix
        return NotImplemented

    def __hash__(self):
        return hash(self.fix)

    def __repr__(self):
        """Convert to formal string, for repr().
        """
        return "%s.%s(%r)" % (self.__class__.__module__,
                              self.__class__.__qualname__,
                              self.fix)

    def utcoffset(self, dt):
        return datetime.timedelta(hours=self.fix)

    def dst(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return str(self.fix)



import datetime

from aerofiles.igc import patterns
from aerofiles.util.timezone import TimeZoneFix


class Writer:
    """
    A writer for the IGC flight log file format.
    """

    REQUIRED_HEADERS = [
        'manufacturer_code',
        'logger_id',
        'date',
        'logger_type',
        'gps_receiver',
    ]

    def __init__(self, fp=None):
        self.fp = fp
        self.fix_extensions = None
        self.k_record_extensions = None

    def format_date(self, date):
        if isinstance(date, datetime.datetime):
            date = date.date()

        if isinstance(date, datetime.date):
            date = date.strftime('%d%m%y')

        if not patterns.DATE.match(date):
            raise ValueError("Invalid date: " + date)

        return date

    def format_time(self, time):
        if isinstance(time, datetime.datetime):
            time = time.time()

        if isinstance(time, datetime.time):
            time = time.strftime('%H%M%S')

        if not patterns.TIME.match(time):
            raise ValueError("Invalid time: " + time)

        return time

    def format_coordinate(self, value, default=None, is_latitude=True):
        if value is None:
            return default

        if is_latitude:
            if not -90 <= value <= 90:
                raise ValueError('Invalid latitude: %s' % value)

            hemisphere = 'S' if value < 0 else 'N'
            format = '%02d%05d%s'

        else:
            if not -180 <= value <= 180:
                raise ValueError('Invalid longitude: %s' % value)

            hemisphere = 'W' if value < 0 else 'E'
            format = '%03d%05d%s'

        value = abs(value)
        degrees = int(value)
        milliminutes = round((value - degrees) * 60000)
        return format % (degrees, milliminutes, hemisphere)

    def format_latitude(self, value):
        return self.format_coordinate(
            value, default='0000000N', is_latitude=True)

    def format_longitude(self, value):
        return self.format_coordinate(
            value, default='00000000E', is_latitude=False)

    def write_line(self, line):
        self.fp.write((line + u'\r\n').encode('ascii', 'replace'))

    def write_record(self, type, record):
        self.write_line(type + record)

    def write_logger_id(self, manufacturer, logger_id, extension=None,
                        validate=True):
        """
        Write the manufacturer and logger id header line::

            writer.write_logger_id('XXX', 'ABC', extension='FLIGHT:1')
            # -> AXXXABCFLIGHT:1

        Some older loggers have decimal logger ids which can be written like
        this::

            writer.write_logger_id('FIL', '13961', validate=False)
            # -> AFIL13961

        :param manufacturer: the three-letter-code of the manufacturer
        :param logger_id: the logger id as three-letter-code
        :param extension: anything else that should be appended to this header
            (e.g. ``FLIGHT:1``)
        :param validate: whether to validate the manufacturer and logger_id
            three-letter-codes
        """

        if validate:
            if not patterns.MANUFACTURER_CODE.match(manufacturer):
                raise ValueError('Invalid manufacturer code')
            if not patterns.LOGGER_ID.match(logger_id):
                raise ValueError('Invalid logger id')

        record = '%s%s' % (manufacturer, logger_id)
        if extension:
            record = record + extension

        self.write_record('A', record)

    def write_header(self, source, subtype, value,
                     subtype_long=None, value_long=None):
        if source not in ('F', 'O'):
            raise ValueError('Invalid source: %s' % source)

        if not subtype_long:
            record = '%s%s%s' % (source, subtype, value)
        elif not value_long:
            record = '%s%s%s:%s' % (source, subtype, subtype_long, value)
        else:
            record = '%s%s%s%s:%s' % \
                (source, subtype, value, subtype_long, value_long)

        self.write_record('H', record)

    def write_fr_header(self, subtype, value,
                        subtype_long=None, value_long=None):
        self.write_header(
            'F', subtype, value,
            subtype_long=subtype_long, value_long=value_long
        )

    def write_date(self, date):
        """
        Write the date header::

            writer.write_date(datetime.date(2014, 5, 2))
            # -> HFDTE140502

        :param date: a :class:`datetime.date` instance
        """

        self.write_fr_header('DTE', self.format_date(date))

    def write_fix_accuracy(self, accuracy=None):
        """
        Write the GPS fix accuracy header::

            writer.write_fix_accuracy()
            # -> HFFXA500

            writer.write_fix_accuracy(25)
            # -> HFFXA025

        :param accuracy: the estimated GPS fix accuracy in meters (optional)
        """

        if accuracy is None:
            accuracy = 500

        accuracy = int(accuracy)
        if not 0 < accuracy < 1000:
            raise ValueError('Invalid fix accuracy')

        self.write_fr_header('FXA', '%03d' % accuracy)

    def write_pilot(self, pilot):
        """
        Write the pilot declaration header::

            writer.write_pilot('Tobias Bieniek')
            # -> HFPLTPILOTINCHARGE:Tobias Bieniek

        :param pilot: name of the pilot
        """
        self.write_fr_header('PLT', pilot, subtype_long='PILOTINCHARGE')

    def write_copilot(self, copilot):
        """
        Write the copilot declaration header::

            writer.write_copilot('John Doe')
            # -> HFCM2CREW2:John Doe

        :param copilot: name of the copilot
        """
        self.write_fr_header('CM2', copilot, subtype_long='CREW2')

    def write_glider_type(self, glider_type):
        """
        Write the glider type declaration header::

            writer.write_glider_type('Hornet')
            # -> HFGTYGLIDERTYPE:Hornet

        :param glider_type: the glider type (e.g. ``Hornet``)
        """
        self.write_fr_header('GTY', glider_type, subtype_long='GLIDERTYPE')

    def write_glider_id(self, glider_id):
        """
        Write the glider id declaration header::

            writer.write_glider_id('D-4449')
            # -> HFGIDGLIDERID:D-4449

        The glider id is usually the official registration number of the
        airplane. For example:``D-4449`` or ``N116EL``.

        :param glider_id: the glider registration number
        """
        self.write_fr_header('GID', glider_id, subtype_long='GLIDERID')

    def write_gps_datum(self, code=None, gps_datum=None):
        """
        Write the mandatory GPS datum header::

            writer.write_gps_datum()
            # -> HFDTM100GPSDATUM:WGS-1984

            writer.write_gps_datum(33, 'Guam-1963')
            # -> HFDTM033GPSDATUM:Guam-1963

        Note that the default GPS datum is WGS-1984 and you should use that
        unless you have very good reasons against it.

        :param code: the GPS datum code as defined in the IGC file
            specification, section A8
        :param gps_datum: the GPS datum in written form
        """

        if code is None:
            code = 100

        if gps_datum is None:
            gps_datum = 'WGS-1984'

        self.write_fr_header(
            'DTM',
            '%03d' % code,
            subtype_long='GPSDATUM',
            value_long=gps_datum,
        )

    def write_firmware_version(self, firmware_version):
        """
        Write the firmware version header::

            writer.write_firmware_version('6.4')
            # -> HFRFWFIRMWAREVERSION:6.4

        :param firmware_version: the firmware version of the flight recorder
        """
        self.write_fr_header(
            'RFW', firmware_version, subtype_long='FIRMWAREVERSION')

    def write_hardware_version(self, hardware_version):
        """
        Write the hardware version header::

            writer.write_hardware_version('1.2')
            # -> HFRHWHARDWAREVERSION:1.2

        :param hardware_version: the hardware version of the flight recorder
        """
        self.write_fr_header(
            'RHW', hardware_version, subtype_long='HARDWAREVERSION')

    def write_logger_type(self, logger_type):
        """
        Write the extended logger type header::

            writer.write_logger_type('Flarm-IGC')
            # -> HFFTYFRTYPE:Flarm-IGC

        :param logger_type: the extended type information of the flight
            recorder
        """
        self.write_fr_header('FTY', logger_type, subtype_long='FRTYPE')

    def write_gps_receiver(self, gps_receiver):
        """
        Write the GPS receiver header::

            writer.write_gps_receiver('uBLOX LEA-4S-2,16,max9000m')
            # -> HFGPSuBLOX LEA-4S-2,16,max9000m

        :param gps_receiver: the GPS receiver information
        """
        self.write_fr_header('GPS', gps_receiver, subtype_long='RECEIVER')

    def write_pressure_sensor(self, pressure_sensor):
        """
        Write the pressure sensor header::

            writer.write_pressure_sensor('Intersema MS5534B,8191')
            # -> HFPRSPRESSALTSENSOR:Intersema MS5534B,8191

        :param pressure_sensor: the pressure sensor information
        """
        self.write_fr_header(
            'PRS', pressure_sensor, subtype_long='PRESSALTSENSOR')

    def write_competition_id(self, competition_id):
        """
        Write the optional competition id declaration header::

            writer.write_competition_id('TH')
            # -> HFCIDCOMPETITIONID:TH

        :param competition_id: competition id of the glider
        """
        self.write_fr_header(
            'CID', competition_id, subtype_long='COMPETITIONID')

    def write_competition_class(self, competition_class):
        """
        Write the optional competition class declaration header::

            writer.write_competition_class('Club')
            # -> HFCCLCOMPETITIONCLASS:Club

        :param competition_class: competition class of the glider
        """
        self.write_fr_header(
            'CCL', competition_class, subtype_long='COMPETITIONCLASS')

    def write_club(self, club):
        """
        Write the optional club declaration header::

            writer.write_club('LV Aachen')
            # -> HFCLBCLUB:LV Aachen

        :param club: club or organisation for which this flight should be
            scored
        """
        self.write_fr_header('CLB', club, subtype_long='CLUB')

    def write_headers(self, headers):
        """
        Write all the necessary headers in the correct order::

            writer.write_headers({
                'manufacturer_code': 'XCS',
                'logger_id': 'TBX',
                'date': datetime.date(1987, 2, 24),
                'fix_accuracy': 50,
                'pilot': 'Tobias Bieniek',
                'copilot': 'John Doe',
                'glider_type': 'Duo Discus',
                'glider_id': 'D-KKHH',
                'firmware_version': '2.2',
                'hardware_version': '2',
                'logger_type': 'LXNAVIGATION,LX8000F',
                'gps_receiver': 'uBLOX LEA-4S-2,16,max9000m',
                'pressure_sensor': 'INTERSEMA,MS5534A,max10000m',
                'competition_id': '2H',
                'competition_class': 'Doubleseater',
            })

            # -> AXCSTBX
            # -> HFDTE870224
            # -> HFFXA050
            # -> HFPLTPILOTINCHARGE:Tobias Bieniek
            # -> HFCM2CREW2:John Doe
            # -> HFGTYGLIDERTYPE:Duo Discus
            # -> HFGIDGLIDERID:D-KKHH
            # -> HFDTM100GPSDATUM:WGS-1984
            # -> HFRFWFIRMWAREVERSION:2.2
            # -> HFRHWHARDWAREVERSION:2
            # -> HFFTYFRTYPE:LXNAVIGATION,LX8000F
            # -> HFGPSRECEIVER:uBLOX LEA-4S-2,16,max9000m
            # -> HFPRSPRESSALTSENSOR:INTERSEMA,MS5534A,max10000m
            # -> HFCIDCOMPETITIONID:2H
            # -> HFCCLCOMPETITIONCLASS:Doubleseater

        This method will throw a :class:`ValueError` if a mandatory header is
        missing and will fill others up with empty strings if no value was
        given. The optional headers are only written if they are part of the
        specified :class:`dict`.

        .. admonition:: Note

            The use of this method is encouraged compared to calling all the
            other header-writing methods manually!

        :param headers: a :class:`dict` of all the headers that should be
            written.
        """

        for header in self.REQUIRED_HEADERS:
            if header not in headers:
                raise ValueError('%s header missing' % header)

        self.write_logger_id(
            headers['manufacturer_code'],
            headers['logger_id'],
            extension=headers.get('logger_id_extension')
        )

        self.write_date(headers['date'])
        self.write_fix_accuracy(headers.get('fix_accuracy'))

        self.write_pilot(headers.get('pilot', ''))
        if 'copilot' in headers:
            self.write_copilot(headers['copilot'])

        self.write_glider_type(headers.get('glider_type', ''))
        self.write_glider_id(headers.get('glider_id', ''))

        self.write_gps_datum(
            code=headers.get('gps_datum_code'),
            gps_datum=headers.get('gps_datum'),
        )

        self.write_firmware_version(headers.get('firmware_version', ''))
        self.write_hardware_version(headers.get('hardware_version', ''))
        self.write_logger_type(headers['logger_type'])
        self.write_gps_receiver(headers['gps_receiver'])
        self.write_pressure_sensor(headers.get('pressure_sensor', ''))

        if 'competition_id' in headers:
            self.write_competition_id(headers['competition_id'])

        if 'competition_class' in headers:
            self.write_competition_class(headers['competition_class'])

        if 'club' in headers:
            self.write_club(headers['club'])

    def write_extensions(self, type, start_byte, extensions):
        num_extensions = len(extensions)
        if num_extensions >= 100:
            raise ValueError('Too many extensions')

        record = '%02d' % num_extensions
        for extension, length in extensions:
            if not patterns.EXTENSION_CODE.match(extension):
                raise ValueError('Invalid extension: %s' % extension)

            end_byte = start_byte + length - 1
            record += '%02d%02d%s' % (start_byte, end_byte, extension)

            start_byte = start_byte + length

        self.write_record(type, record)

    def write_fix_extensions(self, extensions):
        """
        Write the fix extensions description header::

            writer.write_fix_extensions([('FXA', 3), ('SIU', 2), ('ENL', 3)])
            # -> I033638FXA3940SIU4143ENL

        :param extensions: a list of ``(extension, length)`` tuples
        """
        self.fix_extensions = extensions
        self.write_extensions('I', 36, extensions)

    def write_k_record_extensions(self, extensions):
        """
        Write the K record extensions description header::

            writer.write_k_record_extensions([('HDT', 5)])
            # -> J010812HDT

        Use :meth:`~aerofiles.igc.Writer.write_k_record` to write the
        associated records.

        :param extensions: a list of ``(extension, length)`` tuples
        """
        self.k_record_extensions = extensions
        self.write_extensions('J', 8, extensions)

    def write_task_metadata(
            self, declaration_datetime=None, flight_date=None,
            task_number=None, turnpoints=None, text=None):
        """
        Write the task declaration metadata record::

            writer.write_task_metadata(
                datetime.datetime(2014, 4, 13, 12, 53, 02),
                task_number=42,
                turnpoints=3,
            )
            # -> C140413125302000000004203

        There are sensible defaults in place for all parameters except for the
        ``turnpoints`` parameter. If you don't pass that parameter the method
        will raise a :class:`ValueError`. The other parameter defaults are
        mentioned in the list below.

        :param declaration_datetime: a :class:`datetime.datetime` instance of
            the UTC date and time at the time of declaration (default: current
            date and time)
        :param flight_date: a :class:`datetime.date` instance of the intended
            date of the flight (default: ``000000``, which means "use
            declaration date")
        :param task_number: task number for the flight date or an integer-based
            identifier (default: ``0001``)
        :param turnpoints: the number of turnpoints in the task (not counting
            start and finish points!)
        :param text: optional text to append to the metadata record
        """

        if declaration_datetime is None:
            declaration_datetime = datetime.datetime.now(TimeZoneFix(0))

        if isinstance(declaration_datetime, datetime.datetime):
            declaration_datetime = (
                self.format_date(declaration_datetime) +
                self.format_time(declaration_datetime)
            )
        elif not patterns.DATETIME.match(declaration_datetime):
            raise ValueError('Invalid declaration datetime: %s' %
                             declaration_datetime)

        if flight_date is None:
            flight_date = '000000'
        else:
            flight_date = self.format_date(flight_date)

        if task_number is None:
            task_number = 1
        elif not isinstance(task_number, int):
            raise ValueError('Invalid task number: %s' % task_number)

        if not isinstance(turnpoints, int):
            raise ValueError('Invalid turnpoints: %s' % turnpoints)

        record = '{0}{1}{2:04d}{3:02d}'.format(
            declaration_datetime,
            flight_date,
            task_number,
            turnpoints,
        )

        if text:
            record += text

        self.write_record('C', record)

    def write_task_point(self, latitude=None, longitude=None, text='',
                         distance_min=None, distance_max=None,
                         bearing1=None, bearing2=None):
        """
        Write a task declaration point::

            writer.write_task_point(
                latitude=(51 + 7.345 / 60.),
                longitude=(6 + 24.765 / 60.),
                text='Meiersberg',
            )
            # -> C5107345N00624765EMeiersberg

        If no ``latitude`` or ``longitude`` is passed, the fields will be
        filled with zeros (i.e. unknown coordinates). This however should only
        be used for ``TAKEOFF``  and ``LANDING`` points.

        For area tasks there are some additional parameters that can be used
        to specify the relevant areas::

            writer.write_task_point(
                -(12 + 32.112 / 60.),
                -(178 + .001 / 60.),
                'TURN AREA',
                distance_min=12.0,
                distance_max=32.0,
                bearing1=122.0,
                bearing2=182.0,
            )
            # -> C1232112S17800001W00120000032000122000182000TURN AREA

        :param latitude: latitude of the point (between -90 and 90 degrees)
        :param longitude: longitude of the point (between -180 and 180 degrees)
        :param text: type and/or name of the waypoint (e.g. ``TAKEOFF``,
            ``START``, ``TURN 1``, ``TURN 2``, ``FINISH`` or ``LANDING``)
        """

        latitude = self.format_latitude(latitude)
        longitude = self.format_longitude(longitude)

        record = latitude + longitude

        if None not in [distance_min, distance_max, bearing1, bearing2]:
            record += '%04d' % int(distance_min)
            record += '%03d' % int((distance_min - int(distance_min)) * 1000)
            record += '%04d' % int(distance_max)
            record += '%03d' % int((distance_max - int(distance_max)) * 1000)
            record += '%03d' % int(bearing1)
            record += '%03d' % int((bearing1 - int(bearing1)) * 1000)
            record += '%03d' % int(bearing2)
            record += '%03d' % int((bearing2 - int(bearing2)) * 1000)

        if text:
            record += text

        self.write_record('C', record)

    def write_task_points(self, points):
        """
        Write multiple task declaration points with one call::

            writer.write_task_points([
                (None, None, 'TAKEOFF'),
                (51.40375, 6.41275, 'START'),
                (50.38210, 8.82105, 'TURN 1'),
                (50.59045, 7.03555, 'TURN 2', 0, 32.5, 0, 180),
                (51.40375, 6.41275, 'FINISH'),
                (None, None, 'LANDING'),
            ])
            # -> C0000000N00000000ETAKEOFF
            # -> C5124225N00624765ESTART
            # -> C5022926N00849263ETURN 1
            # -> C5035427N00702133E00000000032500000000180000TURN 2
            # -> C5124225N00624765EFINISH
            # -> C0000000N00000000ELANDING

        see the :meth:`~aerofiles.igc.Writer.write_task_point` method for more
        information.

        :param points: a list of ``(latitude, longitude, text)`` tuples. use
            ``(latitude, longitude, text, distance_min, distance_max, bearing1,
            bearing2)`` tuples for area task points.
        """
        for args in points:
            if len(args) not in [3, 7]:
                raise ValueError('Invalid number of task point tuple items')

            self.write_task_point(*args)

    def write_security(self, security, bytes_per_line=75):
        """
        Write the security signature::

            writer.write_security('ABCDEF')
            # -> GABCDEF

        If a signature of more than 75 bytes is used the G record will be
        broken into multiple lines according to the IGC file specification.
        This rule can be configured with the ``bytes_per_line`` parameter if
        necessary.

        :param security: the security signature
        :param bytes_per_line: the maximum number of bytes per line
            (default: ``75``)
        """

        for start in range(0, len(security), bytes_per_line):
            self.write_record('G', security[start:start + bytes_per_line])

    def write_fix(self, time=None, latitude=None, longitude=None, valid=False,
                  pressure_alt=None, gps_alt=None, extensions=None):
        """
        Write a fix record::

            writer.write_fix(
                datetime.time(12, 34, 56),
                latitude=51.40375,
                longitude=6.41275,
                valid=True,
                pressure_alt=1234,
                gps_alt=1432,
            )
            # -> B1234565124225N00624765EA0123401432

        :param time: UTC time of the fix record (default:
            :meth:`~datetime.datetime.utcnow`)
        :param latitude: longitude of the last GPS fix
        :param longitude: latitude of the last GPS fix
        :param valid: ``True`` if the current GPS fix is 3D
        :param pressure_alt: altitude to the ICAO ISA above the 1013.25 hPa
            sea level datum
        :param gps_alt: altitude above the WGS84 ellipsoid
        :param extensions: a list of extension values according to previous
            declaration through
            :meth:`~aerofiles.igc.Writer.write_fix_extensions`
        """

        if time is None:
            time = datetime.datetime.now(TimeZoneFix(0))

        record = self.format_time(time)
        record += self.format_latitude(latitude)
        record += self.format_longitude(longitude)
        record += 'A' if valid else 'V'
        record += '%05d' % (pressure_alt or 0)
        record += '%05d' % (gps_alt or 0)

        if self.fix_extensions or extensions:
            if not (isinstance(extensions, list) and
                    isinstance(self.fix_extensions, list)):
                raise ValueError('Invalid extensions list')

            if len(extensions) != len(self.fix_extensions):
                raise ValueError(
                    'Number of extensions does not match declaration')

            for type_length, value in zip(self.fix_extensions, extensions):
                length = type_length[1]

                if isinstance(value, (int, float)):
                    value = ('%0' + str(length) + 'd') % value

                if len(value) != length:
                    raise ValueError('Extension value has wrong length')

                record += value

        self.write_record('B', record)

    def write_event(self, *args):
        """
        Write an event record::

            writer.write_event(datetime.time(12, 34, 56), 'PEV')
            # -> B123456PEV

            writer.write_event(datetime.time(12, 34, 56), 'PEV', 'Some Text')
            # -> B123456PEVSome Text

            writer.write_event('PEV')  # uses utcnow()
            # -> B121503PEV

        :param time: UTC time of the fix record (default:
            :meth:`~datetime.datetime.utcnow`)
        :param code: event type as three-letter-code
        :param text: additional text describing the event (optional)
        """

        num_args = len(args)
        if not (1 <= num_args <= 3):
            raise ValueError('Invalid number of parameters received')

        if num_args == 3:
            time, code, text = args
        elif num_args == 1:
            code = args[0]
            time = text = None
        elif isinstance(args[0], (datetime.time, datetime.datetime)):
            time, code = args
            text = None
        else:
            code, text = args
            time = None

        if time is None:
            time = datetime.datetime.now(TimeZoneFix(0))

        if not patterns.THREE_LETTER_CODE.match(code):
            raise ValueError('Invalid event code')

        record = self.format_time(time)
        record += code
        if text:
            record += text

        self.write_record('E', record)

    def write_satellites(self, *args):
        """
        Write a satellite constellation record::

            writer.write_satellites(datetime.time(12, 34, 56), [1, 2, 5, 22])
            # -> F12345601020522

        :param time: UTC time of the satellite constellation record (default:
            :meth:`~datetime.datetime.utcnow`)
        :param satellites: a list of satellite IDs as either two-character
            strings or integers below 100
        """

        num_args = len(args)
        if num_args not in (1, 2):
            raise ValueError('Invalid number of parameters received')

        if num_args == 1:
            satellites = args[0]
            time = None
        else:
            time, satellites = args

        if time is None:
            time = datetime.datetime.now(TimeZoneFix(0))

        record = self.format_time(time)

        for satellite in satellites:
            if isinstance(satellite, int):
                satellite = '%02d' % satellite

            if len(satellite) != 2:
                raise ValueError('Invalid satellite ID')

            record += satellite

        self.write_record('F', record)

    def write_k_record(self, *args):
        """
        Write a K record::

            writer.write_k_record_extensions([
                ('FXA', 3), ('SIU', 2), ('ENL', 3),
            ])

            writer.write_k_record(datetime.time(2, 3, 4), ['023', 13, 2])

            # -> J030810FXA1112SIU1315ENL
            # -> K02030402313002

        :param time: UTC time of the k record (default:
            :meth:`~datetime.datetime.utcnow`)
        :param extensions: a list of extension values according to previous
            declaration through
            :meth:`~aerofiles.igc.Writer.write_k_record_extensions`
        """

        num_args = len(args)
        if num_args not in (1, 2):
            raise ValueError('Invalid number of parameters received')

        if num_args == 1:
            extensions = args[0]
            time = None
        else:
            time, extensions = args

        if time is None:
            time = datetime.datetime.utcnow()

        record = self.format_time(time)

        if not (isinstance(extensions, list) and
                isinstance(self.k_record_extensions, list)):
            raise ValueError('Invalid extensions list')

        if len(extensions) != len(self.k_record_extensions):
            raise ValueError(
                'Number of extensions does not match declaration')

        for type_length, value in zip(self.k_record_extensions, extensions):
            length = type_length[1]

            if isinstance(value, (int, float)):
                value = ('%0' + str(length) + 'd') % value

            if len(value) != length:
                raise ValueError('Extension value has wrong length')

            record += value

        self.write_record('K', record)

    def write_comment(self, code, text):
        """
        Write a comment record::

            writer.write_comment('PLT', 'Arrived at the first turnpoint')
            # -> LPLTArrived at the first turnpoint

        :param code: a three-letter-code describing the source of the comment
            (e.g. ``PLT`` for pilot)
        :param text: the text that should be added to the comment
        """

        if not patterns.THREE_LETTER_CODE.match(code):
            raise ValueError('Invalid source')

        self.write_record('L', code + text)



"""
The IGC file format is used by glide computers to record flights.

- `Spec "IGC-approved Flight Recorders - Technical Specification" <https://fai.org/igc-documents>`_
- `Wikipedia <https://en.wikipedia.org/wiki/FAI_Gliding_Commission>`__
"""
# flake8: noqa

from .writer import Writer
from .reader import Reader



import datetime

from .common import SeeYouFileFormat
from .converter import WaypointStyle


class Writer:
    """
    A writer for SeeYou CUP files. Supports waypoints and tasks::

        with open('competition.cup', 'wb') as fp:
            writer = Writer(fp)

    :param fp: file pointer to write to
    :param encoding: the encoding used for the output
    :param file_format: Can be one of the SeeYouFileFormat.

    """

    DIVIDER = u'-----Related Tasks-----'

    DISTANCE_FORMAT_FLOAT = u'%.1f%s'
    DISTANCE_FORMAT_INT = u'%d%s'
    DISTANCE_FORMAT_OTHER = u'%s%s'

    ANGLE_FORMAT_FLOAT = u'%.1f'
    ANGLE_FORMAT_INT = u'%d'
    ANGLE_FORMAT_OTHER = u'%s'

    def __init__(self, fp, encoding='utf-8',
                 file_format=SeeYouFileFormat.ELEVEN):
        self.fp = fp
        self.encoding = encoding
        self.wps = set()
        self.in_task_section = False

        if file_format == SeeYouFileFormat.ELEVEN:
            self.headers = SeeYouFileFormat.HEADER_11
        elif file_format == SeeYouFileFormat.TWELVE:
            self.headers = SeeYouFileFormat.HEADER_12
        elif file_format == SeeYouFileFormat.FORTEEN:
            self.headers = SeeYouFileFormat.HEADER_14

        self.write_fields(self.headers)

    def escape(self, field):
        if not field:
            return ''

        return u'"%s"' % field.replace('\\', '\\\\').replace('"', '\\"')

    def format_coordinate(self, value, is_latitude=True):
        if is_latitude:
            if not -90 <= value <= 90:
                raise ValueError(u'Invalid latitude: %s' % value)

            hemisphere = u'S' if value < 0 else u'N'
            format = u'%02d%06.3f%s'

        else:
            if not -180 <= value <= 180:
                raise ValueError(u'Invalid longitude: %s' % value)

            hemisphere = u'W' if value < 0 else u'E'
            format = u'%03d%06.3f%s'

        value = abs(value)
        degrees = int(value)
        minutes = (value - degrees) * 60
        return format % (degrees, minutes, hemisphere)

    def format_latitude(self, value):
        return self.format_coordinate(value, is_latitude=True)

    def format_longitude(self, value):
        return self.format_coordinate(value, is_latitude=False)

    def format_angle(self, angle):
        if angle is None or angle == '':
            return u''

        if isinstance(angle, float):
            return self.ANGLE_FORMAT_FLOAT % angle
        elif isinstance(angle, int):
            return self.ANGLE_FORMAT_INT % angle
        else:
            return self.ANGLE_FORMAT_OTHER % angle

    def format_distance(self, distance):
        if distance is None or distance == '':
            return u''

        if isinstance(distance, tuple):
            unit = distance[1]
            distance = distance[0]
        else:
            unit = u'm'

        if isinstance(distance, float):
            return self.DISTANCE_FORMAT_FLOAT % (distance, unit)
        elif isinstance(distance, int):
            return self.DISTANCE_FORMAT_INT % (distance, unit)
        else:
            return self.DISTANCE_FORMAT_OTHER % (distance, unit)

    def format_pics(self, pics):
        if pics is None or pics == []:
            return u''
        return self.escape(';'.join(pics))

    def format_time(self, time):
        if isinstance(time, datetime.datetime):
            time = time.time()

        if isinstance(time, datetime.time):
            time = time.strftime(u'%H:%M:%S')

        return time

    def format_timedelta(self, timedelta):
        if isinstance(timedelta, datetime.timedelta):
            hours, remainder = divmod(timedelta.seconds, 3600)
            minutes, seconds = divmod(remainder, 60)
            timedelta = u'%02d:%02d:%02d' % (hours, minutes, seconds)

        return timedelta

    def write_line(self, line=u''):
        self.fp.write((line + u'\r\n').encode(self.encoding))

    def write_fields(self, fields):
        self.write_line(u','.join(fields))

    def set_field(self, field, key, value):
        if key in self.headers:
            field[self.headers.index(key)] = value
        else:
            raise RuntimeError("Writing value to non existing column '%s'. Check for correct SeeYouFileFormat when creating Writer." % key)

    def write_waypoint(
            self, name, shortname, country, latitude, longitude, elevation=u'',
            style=WaypointStyle.NORMAL, runway_direction=u'', runway_length=u'',
            frequency=u'', description=u'', runway_width=u'', userdata=u'', pics=u''):

        """
        Write a waypoint::

            writer.write_waypoint(
                'Meiersberg',
                'MEIER',
                'DE',
                (51 + 7.345 / 60.),
                (6 + 24.765 / 60.),
            )
            # -> "Meiersberg","MEIER",DE,5107.345N,00624.765E,,1,,,,

        :param name: name of the waypoint (must not be empty)
        :param shortname: short name for depending GPS devices
        :param country: IANA top level domain country code (see
            http://www.iana.org/cctld/cctld-whois.htm)
        :param latitude: latitude of the point (between -90 and 90 degrees)
        :param longitude: longitude of the point (between -180 and 180 degrees)
        :param elevation: elevation of the waypoint in meters or as
            ``(elevation, unit)`` tuple
        :param style: the waypoint type (see official specification for the
            list of valid styles, defaults to "Normal")
        :param runway_direction: heading of the runway in degrees if the
            waypoint is landable
        :param runway_length: length of the runway in meters or as ``(length,
            unit)`` tuple if the waypoint is landable
        :param frequency: radio frequency of the airport
        :param description: optional description of the waypoint (no length
            limit)
        :param runway_width: width of the runway in meters or as ``(width,
            unit)`` tuple if the waypoint is landable
        :param userdata: arbitrary string with user data (no length
            limit)
        :param pics: list of filenames of pictures

        """

        if self.in_task_section:
            raise RuntimeError(u'Waypoints must be written before any tasks')

        if not name:
            raise ValueError(u'Waypoint name must not be empty')

        fields = [""] * len(self.headers)
        self.set_field(fields, 'name', self.escape(name))
        self.set_field(fields, 'code', self.escape(shortname))
        self.set_field(fields, 'country', country)
        self.set_field(fields, 'lat', self.format_latitude(latitude))
        self.set_field(fields, 'lon', self.format_longitude(longitude))
        self.set_field(fields, 'elev', self.format_distance(elevation))
        self.set_field(fields, 'style', str(style))
        self.set_field(fields, 'rwdir', str(runway_direction))
        self.set_field(fields, 'rwlen', self.format_distance(runway_length))
        if runway_width:
            self.set_field(fields, 'rwwidth', self.format_distance(runway_width))
        self.set_field(fields, 'freq', self.escape(frequency))
        self.set_field(fields, 'desc', self.escape(description))
        if userdata:
            self.set_field(fields, 'userdata', self.escape(userdata))
        if pics:
            self.set_field(fields, 'pics', self.format_pics(pics))

        self.write_fields(fields)

        self.wps.add(name)

    def write_task(self, description, waypoints):

        """
        Write a task definition::

            writer.write_task('500 km FAI', [
                'MEIER',
                'BRILO',
                'AILER',
                'MEIER',
            ])
            # -> "500 km FAI","MEIER","BRILO","AILER","MEIER"

        Make sure that the referenced waypoints have been written with
        :meth:`~aerofiles.seeyou.Writer.write_waypoint` before writing the
        task. The task section divider will be written to automatically when
        :meth:`~aerofiles.seeyou.Writer.write_task` is called the first time.
        After the first task is written
        :meth:`~aerofiles.seeyou.Writer.write_waypoint` must not be called
        anymore.

        :param description: description of the task (may be blank)
        :param waypoints: list of waypoints in the task (names must match the
            long names of previously written waypoints)
        """

        if not self.in_task_section:
            self.write_line()
            self.write_line(self.DIVIDER)
            self.in_task_section = True

        fields = [self.escape(description)]

        for waypoint in waypoints:
            if waypoint not in self.wps:
                raise ValueError(u'Waypoint "%s" was not found' % waypoint)

            fields.append(self.escape(waypoint))

        self.write_fields(fields)

    def write_task_options(self, **kw):

        """
        Write an options line for a task definition::

            writer.write_task_options(
                start_time=time(12, 34, 56),
                task_time=timedelta(hours=1, minutes=45, seconds=12),
                waypoint_distance=False,
                distance_tolerance=(0.7, 'km'),
                altitude_tolerance=300.0,
            )
            # -> Options,NoStart=12:34:56,TaskTime=01:45:12,WpDis=False,NearDis=0.7km,NearAlt=300.0m

        :param start_time: opening time of the start line as
            :class:`datetime.time` or string
        :param task_time: designated time for the task as
            :class:`datetime.timedelta` or string
        :param waypoint_distance: task distance calculation (``False``: use
            fixes, ``True``: use waypoints)
        :param distance_tolerance: distance tolerance in meters or as
            ``(distance, unit)`` tuple
        :param altitude_tolerance: altitude tolerance in meters or as
            ``(distance, unit)`` tuple
        :param min_distance: "uncompleted leg (``False``: calculate maximum
            distance from last observation zone)"
        :param random_order: if ``True``, then Random order of waypoints is
            checked
        :param max_points: maximum number of points
        :param before_points: number of mandatory waypoints at the beginning.
            ``1`` means start line only, ``2`` means start line plus first
            point in task sequence (Task line).
        :param after_points: number of mandatory waypoints at the end. ``1``
            means finish line only, ``2`` means finish line and one point
            before finish in task sequence (Task line).
        :param bonus: bonus for crossing the finish line
        """

        if not self.in_task_section:
            raise RuntimeError(
                u'Task options have to be written in task section')

        fields = ['Options']

        if 'start_time' in kw:
            fields.append(u'NoStart=' + self.format_time(kw['start_time']))

        if 'task_time' in kw:
            fields.append(u'TaskTime=' + self.format_timedelta(kw['task_time']))

        if 'waypoint_distance' in kw:
            fields.append(u'WpDis=%s' % kw['waypoint_distance'])

        if 'distance_tolerance' in kw:
            fields.append(
                u'NearDis=' + self.format_distance(kw['distance_tolerance']))

        if 'altitude_tolerance' in kw:
            fields.append(
                u'NearAlt=' + self.format_distance(kw['altitude_tolerance']))

        if 'min_distance' in kw:
            fields.append(u'MinDis=%s' % kw['min_distance'])

        if 'random_order' in kw:
            fields.append(u'RandomOrder=%s' % kw['random_order'])

        if 'max_points' in kw:
            fields.append(u'MaxPts=%d' % kw['max_points'])

        if 'before_points' in kw:
            fields.append(u'BeforePts=%d' % kw['before_points'])

        if 'after_points' in kw:
            fields.append(u'AfterPts=%d' % kw['after_points'])

        if 'bonus' in kw:
            fields.append(u'Bonus=%d' % kw['bonus'])

        self.write_fields(fields)

    def write_observation_zone(self, num, **kw):

        """
        Write observation zone information for a taskpoint::

            writer.write_task_options(
                start_time=time(12, 34, 56),
                task_time=timedelta(hours=1, minutes=45, seconds=12),
                waypoint_distance=False,
                distance_tolerance=(0.7, 'km'),
                altitude_tolerance=300.0,
            )
            # -> Options,NoStart=12:34:56,TaskTime=01:45:12,WpDis=False,NearDis=0.7km,NearAlt=300.0m

        :param num: consecutive number of a waypoint (``0``: Start)
        :param style: direction (``0``: Fixed value, ``1``: Symmetrical, ``2``:
            To next point, ``3``: To previous point, ``4``: To start point
        :param radius: radius 1 in meter or as ``(radius, unit)`` tuple
        :param angle: angle 1 in degrees
        :param radius2: radius 2 in meter or as ``(radius, unit)`` tuple
        :param angle 2: angle 2 in degrees
        :param angle12: angle 12 in degress
        :param line: should be ``True`` if start or finish line
        """

        if not self.in_task_section:
            raise RuntimeError(
                u'Observation zones have to be written in task section')

        fields = [u'ObsZone=%d' % num]

        if 'style' in kw:
            fields.append(u'Style=%d' % kw['style'])

        if 'radius' in kw:
            fields.append(u'R1=' + self.format_distance(kw['radius']))

        if 'angle' in kw:
            fields.append(u'A1=' + self.format_angle(kw['angle']))

        if 'radius2' in kw:
            fields.append(u'R2=' + self.format_distance(kw['radius2']))

        if 'angle2' in kw:
            fields.append(u'A2=' + self.format_angle(kw['angle2']))

        if 'angle12' in kw:
            fields.append(u'A12=' + self.format_angle(kw['angle12']))

        if 'line' in kw:
            fields.append(u'Line=' + ('1' if kw['line'] else '0'))

        self.write_fields(fields)



import os

from setuptools import find_packages, setup

GITHUB_URL = 'https://github.com/Turbo87/aerofiles/'


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()


setup(
    name='aerofiles',
    version='1.4.3',
    description='waypoint, task, tracklog readers and writers for aviation',
    long_description=read('README.rst'),
    url=GITHUB_URL,
    license='MIT',
    author='Tobias Bieniek',
    author_email='tobias.bieniek@gmx.de',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: GIS',
    ],
    packages=find_packages(exclude=['tests*']),
)
